@@ -25,8 +25,6 @@ if sys.version > '3':
     _bchr = lambda x: bytes([x])
     _bord = lambda x: x
 
-import array
-import copy
 import struct
 
 import bitcoin.core
@@ -230,6 +228,7 @@ OP_CHECKMULTISIGVERIFY = CScriptOp(0xaf)
 # expansion
 OP_NOP1 = CScriptOp(0xb0)
 OP_NOP2 = CScriptOp(0xb1)
+OP_CHECKLOCKTIMEVERIFY = OP_NOP2
 OP_NOP3 = CScriptOp(0xb2)
 OP_NOP4 = CScriptOp(0xb3)
 OP_NOP5 = CScriptOp(0xb4)
@@ -351,6 +350,7 @@ OPCODE_NAMES.update({
     OP_CHECKMULTISIGVERIFY: 'OP_CHECKMULTISIGVERIFY',
     OP_NOP1: 'OP_NOP1',
     OP_NOP2: 'OP_NOP2',
+    OP_CHECKLOCKTIMEVERIFY: 'OP_CHECKLOCKTIMEVERIFY',
     OP_NOP3: 'OP_NOP3',
     OP_NOP4: 'OP_NOP4',
     OP_NOP5: 'OP_NOP5',
@@ -470,6 +470,7 @@ OPCODES_BY_NAME = {
     'OP_CHECKMULTISIGVERIFY': OP_CHECKMULTISIGVERIFY,
     'OP_NOP1': OP_NOP1,
     'OP_NOP2': OP_NOP2,
+    'OP_CHECKLOCKTIMEVERIFY': OP_CHECKLOCKTIMEVERIFY,
     'OP_NOP3': OP_NOP3,
     'OP_NOP4': OP_NOP4,
     'OP_NOP5': OP_NOP5,
@@ -620,7 +621,9 @@ class CScript(bytes):
         PUSHDATA encodings.
         """
         for (opcode, data, sop_idx) in self.raw_iter():
-            if data is not None:
+            if opcode == 0:
+                yield 0
+            elif data is not None:
                 yield data
             else:
                 opcode = CScriptOp(opcode)
@@ -771,11 +774,6 @@ class CScript(bytes):
         return n
 
 
-SCRIPT_VERIFY_P2SH = object()
-SCRIPT_VERIFY_STRICTENC = object()
-SCRIPT_VERIFY_EVEN_S = object()
-SCRIPT_VERIFY_NOCACHE = object()
-
 SIGHASH_ALL = 1
 SIGHASH_NONE = 2
 SIGHASH_SINGLE = 3
@@ -1030,6 +1028,7 @@ __all__ = (
         'OP_CHECKMULTISIGVERIFY',
         'OP_NOP1',
         'OP_NOP2',
+        'OP_CHECKLOCKTIMEVERIFY',
         'OP_NOP3',
         'OP_NOP4',
         'OP_NOP5',
@@ -1049,10 +1048,6 @@ __all__ = (
         'CScriptInvalidError',
         'CScriptTruncatedPushDataError',
         'CScript',
-        'SCRIPT_VERIFY_P2SH',
-        'SCRIPT_VERIFY_STRICTENC',
-        'SCRIPT_VERIFY_EVEN_S',
-        'SCRIPT_VERIFY_NOCACHE',
         'SIGHASH_ALL',
         'SIGHASH_NONE',
         'SIGHASH_SINGLE',
